/* Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * Copyright 2014 Cloud Linux Zug GmbH
 *
 * Licensed under CLOUD LINUX LICENSE AGREEMENT
 * http://cloudlinux.com/docs/LICENSE.TXT
 *
 * The part the support library for lsapi & proxy_lsapi Apache modules
 * author Alexander Demeshko <ademeshko@cloudlinux.com>
 *
 */
 
#include "lsphp.h"
#include <signal.h>

FILE *file_log = NULL;
 
void close_log(void)
{
    if (file_log != NULL) {
        /*
         * ask fcntl(2) to set the FD_CLOEXEC flag on the log file,
         * so it'll be automagically closed if the exec() call succeeds.
         */
        fflush(file_log);
        setbuf(file_log, NULL);
        if ((fcntl(fileno(file_log), F_SETFD, FD_CLOEXEC) == -1)) {
            log_err("error: can't set close-on-exec flag");
            exit(122);
        }
    }
}
 
void err_output(int is_error, const char *fmt, va_list ap)
{
    time_t timevar;
    struct tm *lt;

    if (!file_log) {
#if defined(_LARGEFILE64_SOURCE) && HAVE_FOPEN64
        if ((file_log = fopen64(AP_LOG_EXEC, "a")) == NULL) {
#else
        if ((file_log = fopen(AP_LOG_EXEC, "a")) == NULL) {
#endif
            fprintf(stderr, AP_PROGRAM_NAME " failure: could not open log file\n");
            perror("fopen");
            exit(1);
        }
    }

    if (is_error) {
        fprintf(stderr, AP_PROGRAM_NAME " policy violation: see its log for more "
                        "details\n");
    }

    time(&timevar);
    lt = localtime(&timevar);

    fprintf(file_log, "[%d-%.2d-%.2d %.2d:%.2d:%.2d]: ",
            lt->tm_year + 1900, lt->tm_mon + 1, lt->tm_mday,
            lt->tm_hour, lt->tm_min, lt->tm_sec);

    vfprintf(file_log, fmt, ap);

    fflush(file_log);
    return;
}

void log_err(const char *fmt,...)
{
    va_list ap;

    va_start(ap, fmt);
    err_output(1, fmt, ap); /* 1 == is_error */
    va_end(ap);
    return;
}

void log_no_err(const char *fmt,...)
{
    va_list ap;

    va_start(ap, fmt);
    err_output(0, fmt, ap); /* 0 == !is_error */
    va_end(ap);
    return;
}

/*
static void reopen_stderr(const char *target_uname)
{
   char pathbuf[256];
   snprintf(pathbuf, sizeof pathbuf, "%s.%s", AP_LOG_EXEC, target_uname);
   
    
   int fd = open(pathbuf, O_CREAT | O_APPEND | O_RDWR, 0666);
   //log_no_err("open(%s) ret %d: %d\n", pathbuf, fd, errno);
   if (fd >= 0)
   {
       int rc = dup2(2, fd);
       //log_no_err("dup2 ret %d: %d\n", rc, errno);
   }
}
*/

static const char *const safe_env_lst[] =
{
    /* variable name starts with */
    "LSAPI_",
    "PHP_LSAPI_",

    /* variable name is */
    "PHPRC=",
    NULL
};

extern char **environ;

void clean_env(void)
{
    char pathbuf[512];
    char **cleanenv;
    char **ep;
    int cidx = 0;
    int idx;

    /* While cleaning the environment, the environment should be clean.
     * (e.g. malloc() may get the name of a file for writing debugging info.
     * Bad news if MALLOC_DEBUG_FILE is set to /etc/passwd.  Sprintf() may be
     * susceptible to bad locale settings....)
     * (from PR 2790)
     */
    char **envp = environ;
    char *empty_ptr = NULL;

    environ = &empty_ptr; /* VERY safe environment */

    if ((cleanenv = (char **) calloc(AP_ENVBUF, sizeof(char *))) == NULL) {
        log_err("failed to malloc memory for environment\n");
        exit(123);
    }

    sprintf(pathbuf, "PATH=%s", AP_SAFE_PATH);
    cleanenv[cidx] = strdup(pathbuf);
    if (cleanenv[cidx] == NULL) {
        log_err("failed to malloc memory for environment\n");
        exit(124);
    }
    cidx++;

    for (ep = envp; *ep && cidx < AP_ENVBUF-1; ep++) {
        for (idx = 0; safe_env_lst[idx]; idx++) {
            if (!strncmp(*ep, safe_env_lst[idx],
                         strlen(safe_env_lst[idx]))) {
                cleanenv[cidx] = *ep;
                cidx++;
                break;
            }
        }
    }

    cleanenv[cidx] = NULL;

    environ = cleanenv;
}

void fork_itself(void) {
    pid_t p = fork();
    if(p > 0) {
        exit(0);
    }
    
    if(p < 0) {
        log_no_err("fork failed: %d(%s) - continue\n", errno, strerror(errno));
    }
}

void lve_jail(struct passwd *pw) 
{
    void *lib_handle;
    lib_handle = dlopen("liblve.so.0", RTLD_LAZY);
    if (lib_handle) {
        char *error;
        char error_msg[8192];
        int (*jail)(struct passwd *, char*) = dlsym(lib_handle, "jail");
        if ((error = dlerror()) != NULL) {
            log_err("failed to init LVE library %s\n", error);
            exit(130);
        }
        int result = jail(pw, error_msg);
        if (result < 0) {
            log_err("SecureLVE jail error %s\n", error_msg);
            exit(131);
        }
    }
}

void set_signals(void) 
{
    sigset_t sig_set;
    sigemptyset(&sig_set);
    sigprocmask(SIG_SETMASK, &sig_set, NULL);
}
